"""
Running functional tests against a router - if the router is present.
"""

import pytest

from ..core.exceptions import (
    FritzConnectionException,
    FritzServiceError,
)
from ..core.fritzconnection import FritzConnection
from ..lib.fritzcall import FritzCall
from ..lib.fritzhomeauto import FritzHomeAutomation
from ..lib.fritzhosts import FritzHosts
from ..lib.fritzphonebook import FritzPhonebook
from ..lib.fritzstatus import FritzStatus
from ..lib.fritzwlan import FritzWLAN


TIMEOUT = 2.0  # give older models some time


@pytest.fixture(scope="module")
def get_fc_instance():
    return FritzConnection(timeout=TIMEOUT)


@pytest.fixture(scope="module")
def get_fc_tls_instance():
    return FritzConnection(timeout=TIMEOUT, use_tls=True)


@pytest.mark.routertest
@pytest.mark.parametrize("use_tls", [False, True])
def test_access_model_name(use_tls):
    """
    Check whether description files are accessible.
    In this case there is some modelname available (a string).
    """
    fc = FritzConnection(use_tls=use_tls)
    # on success the modelname should be a string:
    assert isinstance(fc.modelname, str)


@pytest.mark.routertest
@pytest.mark.parametrize("use_tls", [False, True])
def test_soap_access(use_tls, get_fc_instance, get_fc_tls_instance):
    """
    Test whether a soap access returns successful. The Service
    'DeviceInfo1' with the action 'GetInfo' should be provided by all
    router models – including repeaters - and provides access to the
    model name.
    The service must be accessible without a password. As the model is unknown, the service should be available on all models. In practice that's not the case because routers and repeaters will have different service sets. So this will test a soap-call if
    """
    fc = FritzConnection(use_tls=use_tls)
    try:
        info = fc.call_action('DeviceInfo1', 'GetInfo')
    except FritzConnectionException:
        # successful call but access not allowed
        # (password required)
        # so at least the soap request has worked:
        assert True
    except FritzServiceError:
        # this should not happen with the known product range.
        # skip this:
        pass
    else:
        # on success the modelname should be a string:
        assert isinstance(info['NewModelName'], str)
    try:
        info = fc.call_action('WANIPConn1', 'GetStatusInfo')
    except FritzServiceError:
        # some devices (like repeaters) may not have this service.
        # skip this:
        pass
    else:
        assert isinstance(info['NewConnectionStatus'], str)
    # do at least one assert:
    assert True


@pytest.mark.routertest
@pytest.mark.parametrize(
    "cls, use_tls", [
        (FritzCall, False),
        (FritzCall, True),
        (FritzHomeAutomation, False),
        (FritzHomeAutomation, True),
        (FritzHosts, False),
        (FritzHosts, True),
        (FritzPhonebook, False),
        (FritzPhonebook, True),
        (FritzStatus, False),
        (FritzStatus, True),
        (FritzWLAN, False),
        (FritzWLAN, True),
    ])
def test_library_api_arguments(cls, use_tls):
    obj = cls(timeout=TIMEOUT, use_tls=use_tls)
    # on success the modelname should be a string:
    assert isinstance(obj.modelname, str)


@pytest.mark.routertest
@pytest.mark.parametrize(
    "cls", [
        FritzCall,
        FritzHomeAutomation,
        FritzHosts,
        FritzPhonebook,
        FritzStatus,
        FritzWLAN
    ])
def test_init_cls_with_instance(cls, get_fc_instance):
    obj = cls(fc=get_fc_instance)
    assert isinstance(obj.modelname, str)


@pytest.mark.routertest
@pytest.mark.parametrize(
    "cls", [
        FritzCall,
        FritzHomeAutomation,
        FritzHosts,
        FritzPhonebook,
        FritzStatus,
        FritzWLAN
    ])
def test_init_cls_with_tls_instance(cls, get_fc_tls_instance):
    obj = cls(fc=get_fc_tls_instance)
    assert isinstance(obj.modelname, str)



import pytest

from ..lib.fritztools import (
    ArgumentNamespace,
    byte_formatter,
    format_num,
)


@pytest.mark.parametrize(
    "value, result, dimension", [
        (1, 1.0, 'B'),
        (123, 123.0, 'B'),
        (1230, 1.230, 'KB'),
        (12345, 12.345, 'KB'),
        (242981246, 242.981246, 'MB'),
        (24298124612, 24.298124612, 'GB'),
        (2429812461200, 2.4298124612, 'TB'),
        (42e3, 42.0, 'KB'),
        (42e6, 42.0, 'MB'),
        (42e9, 42.0, 'GB'),
        (42e12, 42.0, 'TB'),
        (42e15, 42.0, 'PB'),
        (42e18, 42000.0, 'PB'),
        (1.0, 1.0, 'B'),
        (0.1, 0, 'B'),
        (0.01, 0, 'B'),
        (0.001, 0, 'B'),
        (0, 0, 'B'),
        (-10, 10, 'B'),
    ]
)
def test_byte_formatter(value, result, dimension):
    num, dim = byte_formatter(value)
    assert num == result
    assert dim == dimension


@pytest.mark.parametrize(
    "num, formated_num", [
        (300, '300.0 B'),
        (2000, '2.0 KB'),
        (3500, '3.5 KB'),
        (32500, '32.5 KB'),
        (242911246, '242.9 MB'),
        (242981246, '243.0 MB'),
        (45e6, '45.0 MB'),
        (45e9, '45.0 GB'),
        (45e12, '45.0 TB'),
        (45e15, '45.0 PB'),
        (45e18, '45000.0 PB'),
    ]
)
def test_format_num(num, formated_num):
    assert formated_num == format_num(num)


def test_format_num_bits():
    result = format_num(1234, unit='bits')
    assert result == '1.2 KBit'


def test_argument_namespace():
    source = {
        'NewManufacturerName': 'AVM',
        'NewManufacturerOUI': '00040E',
        'NewModelName': 'FRITZ!Box 7590',
        'NewDescription': 'FRITZ!Box 7590 154.07.29',
        'NewProductClass': 'AVMFB7590',
        'NewSerialNumber': '989BCB2B93B0',
        'NewSoftwareVersion': '154.07.29',
        'NewHardwareVersion': 'FRITZ!Box 7590',
        'NewSpecVersion': '1.0',
        'NewProvisioningCode': '000.044.004.000',
        'NewUpTime': 9516949,
        'NewDeviceLog': 'long string here ...'
    }
    mapping = {
        "serial_number": "NewSerialNumber",
        "model_name": "NewModelName",
    }
    info = ArgumentNamespace(source, mapping)
    assert info.serial_number == '989BCB2B93B0'
    assert info['model_name'] == 'FRITZ!Box 7590'


@pytest.mark.parametrize(
    "name, expected_result", [
        ("description", "description"),
        ("Description", "description"),
        ("ModelName", "model_name"),
        ("NewUpTime", "new_up_time"),
        ("new_up_time", "new_up_time"),
        ("NewManufacturerOUI", "new_manufacturer_oui"),
    ]
)
def test_argument_namespace_rewrite(name, expected_result):
    result = ArgumentNamespace.rewrite_argument(name, suppress_new=False)
    assert result == expected_result


@pytest.mark.parametrize(
    "name, expected_result", [
        ("NewUpTime", "up_time"),
        ("UpTime", "up_time"),
        ("upTime", "up_time"),
        ("up_time", "up_time"),
        ("uptime", "uptime"),
        ("newuptime", "newuptime"),
        ("Newuptime", "newuptime"),
        ("New_uptime", "uptime"),
        ("NewManufacturerOUI", "manufacturer_oui"),
    ]
)
def test_argument_namespace_rewrite_no_new(name, expected_result):
    result = ArgumentNamespace.rewrite_argument(name)
    assert result == expected_result


@pytest.fixture()
def avm_source():
    source = {
        'NewModelName': 'FRITZ!Box 7590',
        'NewDescription': 'FRITZ!Box 7590 154.07.29',
        'NewProductClass': 'AVMFB7590',
        'NewSerialNumber': '989BCB2B93B0',
    }
    return source


def test_argument_namespace_no_mapping(avm_source):
    """
    In case of a missing mapping, all values from source should get
    transfered to the ArgumentNamespace and the key should get converted
    to snake_case.
    """
    info = ArgumentNamespace(avm_source)
    assert info.model_name == 'FRITZ!Box 7590'
    assert info['model_name'] == 'FRITZ!Box 7590'


def test_argument_namespace_no_mapping_no_suppress(avm_source):
    """
    In case of a missing mapping, all values from source should get
    transfered to the ArgumentNamespace and the key should get converted
    to snake_case.
    """
    info = ArgumentNamespace(avm_source, suppress_new=False)
    assert info.new_serial_number == '989BCB2B93B0'
    assert info['new_serial_number'] == '989BCB2B93B0'
    assert info.new_model_name == 'FRITZ!Box 7590'
    assert info['new_model_name'] == 'FRITZ!Box 7590'


def test_argument_namespace_has_len(avm_source):
    info = ArgumentNamespace(avm_source)
    assert len(info) == len(avm_source)


def test_argument_namespace_assignment(avm_source):
    info = ArgumentNamespace(avm_source)
    info.new_value = 42
    assert info.new_value == 42
    assert len(info) == len(avm_source) + 1
    info["minus"] = -3
    assert info["minus"] == -3
    assert info.minus == -3
    assert len(info) == len(avm_source) + 2


def test_argument_namespace_extract(avm_source):
    extract = "NewModelName", "NewProductClass"
    info = ArgumentNamespace(avm_source, extract=extract)
    assert len(info) == len(extract)
    assert info.model_name == avm_source['NewModelName']
    assert info.product_class == avm_source['NewProductClass']



import os
import pytest

from ..core.utils import get_xml_root
from ..core.processor import HostStorage


@pytest.fixture()
def devicehostlist_source(datadir):
    os.chdir(datadir)
    return get_xml_root('devicehostlist.xml')


@pytest.mark.parametrize(
    "idx, attribute, expected_value", [
        (1, "Index", 1),
        (1, "IPAddress", "192.168.178.54"),
        (1, "MACAddress", "F8:4D:89:7C:0A:32"),
        (1, "Active", True),
        (1, "HostName", "MacBook"),
        (1, "InterfaceType", "802.11"),
        (1, "X_AVM-DE_Port", 0),
        (1, "X_AVM-DE_Speed", 866),
        (1, "X_AVM-DE_UpdateAvailable", False),
        (1, "X_AVM-DE_UpdateSuccessful", "unknown"),
        (1, "X_AVM-DE_InfoURL", None),
        (1, "X_AVM-DE_Model", None),
        (1, "X_AVM-DE_URL", None),
        (1, "X_AVM-DE_Guest", False),
        (1, "X_AVM-DE_VPN", False),
        (1, "X_AVM-DE_WANAccess", "granted"),
        (1, "X_AVM-DE_Disallow", False),

        (2, "Index", 2),
        (2, "IPAddress", "192.168.178.34"),
        (2, "MACAddress", "C2:39:6F:FB:A3:E8"),
        (2, "Active", True),
        (2, "HostName", "fritz.repeater"),
        (2, "InterfaceType", "Ethernet"),
        (2, "X_AVM-DE_Port", 3),
        (2, "X_AVM-DE_Speed", 100),
        (2, "X_AVM-DE_UpdateAvailable", False),
        (2, "X_AVM-DE_UpdateSuccessful", "succeeded"),
        (2, "X_AVM-DE_InfoURL", None),
        (2, "X_AVM-DE_Model", "FRITZ!Repeater 1200"),
        (2, "X_AVM-DE_URL", "http://192.168.178.34"),
        (2, "X_AVM-DE_Guest", False),
        (2, "X_AVM-DE_VPN", False),
        (2, "X_AVM-DE_WANAccess", "granted"),
        (2, "X_AVM-DE_Disallow", False),

        (3, "Index", 3),
        (3, "IPAddress", "192.168.178.49"),
        (3, "MACAddress", "B8:27:EB:D1:DC:06"),
        (3, "Active", True),
        (3, "HostName", "raspberrypi"),
        (3, "InterfaceType", "Ethernet"),
        (3, "X_AVM-DE_Port", 1),
        (3, "X_AVM-DE_Speed", 1000),
        (3, "X_AVM-DE_UpdateAvailable", False),
        (3, "X_AVM-DE_UpdateSuccessful", "unknown"),
        (3, "X_AVM-DE_InfoURL", None),
        (3, "X_AVM-DE_Model", None),
        (3, "X_AVM-DE_URL", None),
        (3, "X_AVM-DE_Guest", False),
        (3, "X_AVM-DE_VPN", False),
        (3, "X_AVM-DE_WANAccess", "granted"),
        (3, "X_AVM-DE_Disallow", False),

        (4, "Index", 4),
        (4, "IPAddress", "192.168.178.32"),
        (4, "MACAddress", "B4:CD:27:37:78:E4"),
        (4, "Active", False),
        (4, "HostName", "someones-HUAWEI"),
        (4, "InterfaceType", "802.11"),
        (4, "X_AVM-DE_Port", 0),
        (4, "X_AVM-DE_Speed", 0),
        (4, "X_AVM-DE_UpdateAvailable", False),
        (4, "X_AVM-DE_UpdateSuccessful", "unknown"),
        (4, "X_AVM-DE_InfoURL", None),
        (4, "X_AVM-DE_Model", None),
        (4, "X_AVM-DE_URL", None),
        (4, "X_AVM-DE_Guest", False),
        (4, "X_AVM-DE_VPN", False),
        (4, "X_AVM-DE_WANAccess", "granted"),
        (4, "X_AVM-DE_Disallow", False),
    ])
def test_readhostlist(idx, attribute, expected_value, devicehostlist_source):
    """
    Check for node-extraction and type conversion.
    """
    storage = HostStorage(devicehostlist_source)
    attrs = storage.hosts_attributes[idx-1]  # list is sorted but host-index is 1 based
    assert attrs[attribute] == expected_value



import queue
import socket
import threading
import time

import pytest

from ..core.fritzmonitor import FritzMonitor, EventReporter


class MockSocket:
    """
    A socket dummy to simulate receiving data and optional timeouts.
    """

    def __init__(self, mock_data=None, timeout=None, raise_connect_timeout=False):
        self.mock_data = mock_data
        self.timeout = timeout
        self.raise_connect_timeout = raise_connect_timeout

    def connect(self, *args):
        if self.raise_connect_timeout:
            raise socket.timeout("mock failed socket.connect")

    def close(self):
        pass

    def recv(self, chunk_size=None):
        if not self.mock_data:
            if self.timeout:
                time.sleep(self.timeout)
            raise socket.timeout("mock timeout")
        if chunk_size is None:
            chunk = self.mock_data.encode("utf-8")
        else:
            chunk = self.mock_data[:chunk_size]
            self.mock_data = self.mock_data[chunk_size:]
        return chunk.encode("utf-8")


class MockReconnectSocket(MockSocket):
    """
    Dummy socket to simulate a single reconnect.
    """

    def __init__(self, mock_data=None):
        self.data_provider = self._returner(mock_data)
        self.connect_called_num = 0

    def connect(self, *args):
        self.connect_called_num += 1

    def recv(self, chuck_size=None):
        try:
            data = next(self.data_provider)
        except StopIteration:
            data = " "
        return data.encode("utf-8")

    @staticmethod
    def _returner(mock_data):
        for data in mock_data:
            yield data


class MockReconnectFailSocket(MockReconnectSocket):
    """
    Dummy socket to simulate multiple faild reconnections.
    """

    def __init__(self, mock_data=None, timeouts=0):
        super().__init__(mock_data)
        self._connector = self.connector(timeouts)

    def connect(self, *args):
        super().connect(*args)
        if not next(self._connector):
            raise socket.timeout("mock reconnect timeout")

    @staticmethod
    def connector(timeouts):
        yield True
        for _ in range(timeouts):
            yield False
        while True:
            yield True


def test_init_fritzmonitor():
    fm = FritzMonitor()
    assert fm.monitor_thread is None
    assert fm.stop_flag.is_set() is False


@pytest.mark.parametrize(
    "data, expected",
    [
        ("some data", None),
        ("some_data\nmore_data", "some_data"),
        ("some\nmore\ndata", "some more"),
        ("2019;10\n11:30\ncall\n123\nmore", "2019;10 11:30 call 123"),
    ],
)
def test_event_reporter(data, expected):
    q = queue.Queue()
    ep = EventReporter(q)
    ep.add(data)
    if expected is None:
        # no data should be in the queue
        assert pytest.raises(queue.Empty, q.get_nowait)
    else:
        for event in expected.split():
            item = q.get_nowait()
            assert item == event


def test_start_stop():
    mock_socket = MockSocket(timeout=0.01)
    fm = FritzMonitor()
    assert fm.monitor_thread is None
    fm.start(sock=mock_socket)
    assert fm.monitor_thread is not None
    assert fm.monitor_thread.is_alive() is True
    thread = fm.monitor_thread
    fm.stop()
    assert thread.is_alive() is False
    assert fm.monitor_thread is None


def test_start_stop_properties():
    mock_socket = MockSocket(timeout=0.01)
    fm = FritzMonitor()
    assert fm.has_monitor_thread is False
    assert fm.is_alive is False
    _ = fm.start(sock=mock_socket)
    assert fm.has_monitor_thread is True
    assert fm.is_alive is True
    fm.stop()
    assert fm.has_monitor_thread is False
    assert fm.is_alive is False


def test_queue_and_threading_instances():
    mock_socket = MockSocket(timeout=0.01)
    fm = FritzMonitor()
    event_queue = fm.start(sock=mock_socket)
    assert isinstance(event_queue, queue.Queue)
    assert isinstance(fm.monitor_thread, threading.Thread)
    fm.stop()


def test_start_twice():
    """
    It is a failure to start a running instance again.
    """
    mock_socket = MockSocket(timeout=0.01)
    fm = FritzMonitor()
    fm.start(sock=mock_socket)
    with pytest.raises(RuntimeError):
        # start running instance again: should raise a RuntimeError
        fm.start(sock=mock_socket)
    fm.stop()
    # but starting now again should work:
    fm.start(sock=mock_socket)
    fm.stop()


def test_failed_connection():
    mock_socket = MockSocket(raise_connect_timeout=True)
    fm = FritzMonitor()
    pytest.raises(OSError, fm.start, sock=mock_socket)


@pytest.mark.parametrize(
    "mock_data, chunk_size, expected_events",
    [
        ("simple string of words\n", None, "simple string of words"),
        ("event;call\nnumber\n12345\n", None, "event;call|number|12345"),
        (
            "CALL;20201031;12;34\nINC;20-10-31;09;76\n",
            None,
            "CALL;20201031;12;34|INC;20-10-31;09;76",
        ),
        ("1\n2\n3\n4\n5\n6\n7\n8\n9\n10\n11\n", None, "1|2|3|4|5|6|7|8|9|10|11"),
        ("simple string of words\n", 2048, "simple string of words"),
        ("event;call\nnumber\n12345\n", 2048, "event;call|number|12345"),
        (
            "CALL;20201031;12;34\nINC;20-10-31;09;76\n",
            2048,
            "CALL;20201031;12;34|INC;20-10-31;09;76",
        ),
        ("1\n2\n3\n4\n5\n6\n7\n8\n9\n10\n11\n", 2048, "1|2|3|4|5|6|7|8|9|10|11"),
        ("simple string of words\n", 20, "simple string of words"),
        ("event;call\nnumber\n12345\n", 20, "event;call|number|12345"),
        (
            "CALL;20201031;12;34\nINC;20-10-31;09;76\n",
            20,
            "CALL;20201031;12;34|INC;20-10-31;09;76",
        ),
        ("1\n2\n3\n4\n5\n6\n7\n8\n9\n10\n11\n", 20, "1|2|3|4|5|6|7|8|9|10|11"),
        ("simple string of words\n", 2, "simple string of words"),
        ("event;call\nnumber\n12345\n", 2, "event;call|number|12345"),
        (
            "CALL;20201031;12;34\nINC;20-10-31;09;76\n",
            2,
            "CALL;20201031;12;34|INC;20-10-31;09;76",
        ),
        ("1\n2\n3\n4\n5\n6\n7\n8\n9\n10\n11\n", 2, "1|2|3|4|5|6|7|8|9|10|11"),
    ],
)
def test_get_events(mock_data, chunk_size, expected_events):
    mock_socket = MockSocket(mock_data=mock_data, timeout=0.01)
    fm = FritzMonitor()
    event_queue = fm.start(sock=mock_socket)
    thread = fm.monitor_thread
    received_event_num = 0
    for n, expected_event in enumerate(expected_events.split("|"), start=1):
        try:
            assert event_queue.get(timeout=0.1) == expected_event
            received_event_num += 1
        except queue.Empty:
            pass
    # check if alle expected events have been received
    assert received_event_num == n
    fm.stop()
    assert thread.is_alive() is False


def test_reconnect():
    data = ["first\n", "", "second\n"]
    mock_socket = MockReconnectSocket(data)
    fm = FritzMonitor()
    event_queue = fm.start(sock=mock_socket, reconnect_delay=0.001)
    for expected in [data[0], data[2]]:
        # should not raise _queue.Empty:
        assert event_queue.get(timeout=0.1) == expected.strip()
    fm.stop()
    assert mock_socket.connect_called_num == 2


@pytest.mark.parametrize(
    "timeouts, tries, expected_result",
    [
        (0, 0, True),
        (0, 1, True),
        (1, 1, False),
        (1, 2, True),
        (4, 5, True),
        (5, 5, False),
        (6, 5, False),
    ],
)
def test_MockReconnectFailSocket(timeouts, tries, expected_result):
    """
    Internal test to check whether the MockReconnectFailSocket class works as expected.
    """
    sock = MockReconnectFailSocket(timeouts=timeouts)
    assert sock.connect_called_num == 0
    sock.connect()
    assert sock.connect_called_num == 1
    result = True  # got connection
    for cycle in range(tries):
        try:
            sock.connect()
        except OSError:
            result = False
        else:
            result = True
            break
        finally:
            assert (
                sock.connect_called_num == cycle + 2
            )  # cycle is zero based plus initional connection
    assert result == expected_result


@pytest.mark.parametrize(
    "timeouts", list(range(6)),
)
def test_get_connected_socket(timeouts):
    socket = MockReconnectFailSocket(timeouts=timeouts)
    fm = FritzMonitor()
    fm.mock_socket = socket
    s = fm._get_connected_socket()  # make initional connection
    assert s is socket
    for _ in range(timeouts):
        with pytest.raises(OSError):
            fm._get_connected_socket()
    s = fm._get_connected_socket()
    assert s is socket


@pytest.mark.parametrize(
    "timeouts, tries, expected_result",
    [
        (0, 0, None),
        (0, 1, True),
        (1, 0, None),
        (1, 1, None),
        (1, 2, True),
        (4, 5, True),
        (5, 5, None),
        (6, 5, None),
    ],
)
def test_reconnect_socket(timeouts, tries, expected_result):
    mock_socket = MockReconnectFailSocket(timeouts=timeouts)
    fm = FritzMonitor()
    fm.mock_socket = mock_socket
    _ = fm._get_connected_socket()  # make initional connection
    result = fm._reconnect_socket(
        max_reconnect_delay=0.001, reconnect_tries=tries
    )
    if expected_result is None:
        assert result is expected_result
    else:
        assert result is mock_socket


def test_shut_down_monitor_thread():
    data = ["first_block\n", "second_block\n", ""]
    mock_socket = MockReconnectFailSocket(data, timeouts=1)
    fm = FritzMonitor()
    q = fm.start(sock=mock_socket, reconnect_delay=0.001, reconnect_tries=0)
    for d in data[:-1]:
        time.sleep(0.1)
        value = q.get(timeout=0.1)
        assert value == d.strip()
    assert fm.is_alive is False
    assert fm.has_monitor_thread is False


@pytest.mark.parametrize(
    "data, timeouts, tries, success",
    [
        (["first\n", "second\n"], 0, 0, True),
        (["first\n", "", "second\n"], 1, 0, False),
        (["first\n", "", "second\n"], 0, 1, True),
        (["first\n", "", "second\n"], 1, 1, False),
        (["first\n", "", "second\n"], 1, 2, True),
        # default for tries: 5
        (["first\n", "", "second\n"], 3, 5, True),
        (["first\n", "", "second\n"], 4, 5, True),
        (["first\n", "", "second\n"], 5, 5, False),
    ],
)
def test_terminate_thread_on_failed_reconnection(
    data, timeouts, tries, success
):
    """
    Check for thread-termination in case reconnection fails.
    If the tries are lower than the timeouts, the the start of the
    monitor has failed and the Thread must terminate itself.
    """
    mock_socket = MockReconnectFailSocket(data, timeouts=timeouts)
    fm = FritzMonitor()
    fm.start(sock=mock_socket, reconnect_delay=0.001, reconnect_tries=tries)
    # give thread some time:
    time.sleep(0.01)
    if success:
        assert fm.is_alive is True
    else:
        assert fm.is_alive is False
        assert fm.monitor_thread is None
    fm.stop()



def test_restart_failed_monitor():
    """
    Check whether a fritzmonitor instance with a lost connection can get
    started again. Starting the same instance twice does (and should)
    not work. See test_start_twice(). But after a failed reconnect (a
    lost connection) the same instance without calling stop()
    """
    socket = MockReconnectFailSocket(
        mock_data=["first\n", "", "second\n"], timeouts=16
    )  # just some timeouts
    fm = FritzMonitor()
    fm.start(
        sock=socket, reconnect_delay=0.001, reconnect_tries=5
    )  # set default explicit for clarity
    # give socket some time to lose connection:
    time.sleep(0.01)
    assert fm.is_alive is False
    assert fm.stop_flag.is_set() is False
    # dont' call stop here!
    # fm.stop()
    socket = MockSocket(timeout=0.01)  # socket not losing connection
    # should not trigger a RuntimeError
    fm.start(
        sock=socket, reconnect_delay=0.001, reconnect_tries=5
    )  # set default explicit for clarity
    assert fm.is_alive is True
    fm.stop()


def test_context_manager():
    sock = MockSocket(timeout=0.01)
    with FritzMonitor() as fm:
        fm.start(sock=sock)
        assert fm.is_alive is True
    assert fm.is_alive is False




import pytest


from ..core.utils import (
    boolean_from_string,
    get_boolean_from_string,
)


@pytest.mark.parametrize(
    "value, expected_result", [
        ("True", True),
        ("TRUE", True),
        ("false", False),
        ("fAlSe", False),
        ("On", True),
        ("Off", False),
        ("1", True),
        ("0", False),
    ]
)
def test_boolean_from_string(value, expected_result):
    """
    Test to get a real boolean and not a string.
    """
    result = boolean_from_string(value)
    assert result == expected_result


@pytest.mark.parametrize(
    "value", [
        "wahr",
        "falsch",
        "l",
        "10",
        "01",
        "",
        "None",
    ]
)
def test_boolean_from_string_valueerror(value):
    with pytest.raises(ValueError):
        boolean_from_string(value)


@pytest.mark.parametrize(
    "value", [
        True,
        False,
        None,
        42,
    ]
)
def test_boolean_from_string_typeerror(value):
    with pytest.raises(AttributeError):
        boolean_from_string(value)


@pytest.mark.parametrize(
    "value, expected_result", [
        ("true", True),
        ("false", False),
        ("wahr", None),
        ("falsch", None),
        ("None", None),
        (42, None),
        ("", None),
        (None, None),
    ]
)
def test_get_boolean_from_string(value, expected_result):
    result = get_boolean_from_string(value)
    assert result == expected_result



"""
tests for serialization and deserialization data from the processor classes.
"""

import json

from fritzconnection.core.processor import (
    Action,
    Argument,
    Description,
    Device,
    Scpd,
    Serializer,
    Service,
    SpecVersion,
    StateVariable,
    SystemVersion,
    ValueRange,
)


# some test data.
# data can be anything but take some real data from an existing router:

DATA_SPEC_MAJOR_VERSION = "1"
DATA_SPEC_MINOR_VERSION = "0"

DATA_SYS_MAJOR = "154"
DATA_SYS_MINOR = "7"
DATA_SYS_PATCH = "29"
DATA_SYS_DISPLAY = "154.07.29"

DATA_ARGUMENT_NAME = "NewManufacturerName"
DATA_ARGUMENT_DIRECTION = "out"
DATA_ARGUMENT_RELATEDSTATEVARIABLE = "ManufacturerName"

DATA_ACTION_NAME = "GetInfo"

DATA_VALUERANGE_MINIMUM = None
DATA_VALUERANGE_MAXIMUM = None
DATA_VALUERANGE_STEP = "1"  # for not having all the default values

DATA_STATEVARIABLE_NAME = "LastConnectionError"
DATA_STATEVARIABLE_DATATYPE = "string"
DATA_STATEVARIABLE_DEFAULTVALUE = "ERROR_NONE"
DATA_STATEVARIABLE_ALLOWED_VALUES = ["ERROR_NONE", "ERROR_ISP_TIME_OUT", "ERROR_COMMAND_ABORTED"]

DATA_DEVICE_DEVICETYPE = "urn:schemas-upnp-org:device:InternetGatewayDevice:1"
DATA_DEVICE_FRIENDLYNAME = "FRITZ!Box 7590"
DATA_DEVICE_MANUFACTURER = "AVM Berlin"
DATA_DEVICE_MANUFACTURERURL = "http://www.avm.de"
DATA_DEVICE_MODELDESCRIPTION = "FRITZ!Box 7590"
DATA_DEVICE_MODELNAME = "FRITZ!Box 7590"
DATA_DEVICE_MODELNUMBER = "avm"
DATA_DEVICE_MODELURL = "http://www.avm.de"
DATA_DEVICE_PRESENTATIONURL = "http://fritz.box"
DATA_DEVICE_UDN = "uuid:75802409-bccb-40e7-8e6c-989BCB2B93B0"

DATA_SERVICE_SERVICETYPE = "urn:dslforum-org:service:DeviceInfo:1"
DATA_SERVICE_SERVICEID = "urn:DeviceInfo-com:serviceId:DeviceInfo1"
DATA_SERVICE_CONTROLURL = "/upnp/control/deviceinfo"
DATA_SERVICE_EVENTSUBURL = "/upnp/control/deviceinfo"
DATA_SERVICE_SCPDURL = "/deviceinfoSCPD.xml"

JSON_RESULT_TEST_SERIALIZE_SPECVERSION = f"""{{"major": "{DATA_SPEC_MAJOR_VERSION}", "minor": "{DATA_SPEC_MINOR_VERSION}"}}"""
JSON_RESULT_TEST_SERIALIZE_SYSTEMVERSION = f"""{{"Buildnumber": null, "Display": "{DATA_SYS_DISPLAY}", "HW": null, "Major": "{DATA_SYS_MAJOR}", "Minor": "{DATA_SYS_MINOR}", "Patch": "{DATA_SYS_PATCH}"}}"""
JSON_RESULT_TEST_SERIALIZE_ARGUMENT = f"""{{"direction": "{DATA_ARGUMENT_DIRECTION}", "name": "{DATA_ARGUMENT_NAME}", "relatedStateVariable": "{DATA_ARGUMENT_RELATEDSTATEVARIABLE}"}}"""
JSON_RESULT_TEST_SERIALIZE_VALUERANGE = f"""{{"maximum": null, "minimum": null, "step": "{DATA_VALUERANGE_STEP}"}}"""
JSON_RESULT_TEST_SERIALIZE_ACTION = f"""{{"name": "{DATA_ACTION_NAME}", "arguments": [{JSON_RESULT_TEST_SERIALIZE_ARGUMENT}, {JSON_RESULT_TEST_SERIALIZE_ARGUMENT}]}}"""

# weird but f-strings substitute lists of strings different than json.dumps:
JSON_RESULT_TEST_SERIALIZE_STATEVARIABLE = f"""{{"attributes": {{"allowed_values": """ + json.dumps(DATA_STATEVARIABLE_ALLOWED_VALUES) + f""", "dataType": "{DATA_STATEVARIABLE_DATATYPE}", "defaultValue": "{DATA_STATEVARIABLE_DEFAULTVALUE}", "name": "{DATA_STATEVARIABLE_NAME}"}}, "allowedValueRange": {JSON_RESULT_TEST_SERIALIZE_VALUERANGE}}}"""  # noqa: F541
JSON_RESULT_TEST_SERIALIZE_SCPD = f"""{{"actions": [{JSON_RESULT_TEST_SERIALIZE_ACTION}, {JSON_RESULT_TEST_SERIALIZE_ACTION}], "specVersion": {JSON_RESULT_TEST_SERIALIZE_SPECVERSION}, "state_variables": [{JSON_RESULT_TEST_SERIALIZE_STATEVARIABLE}, {JSON_RESULT_TEST_SERIALIZE_STATEVARIABLE}]}}"""
JSON_RESULT_TEST_SERIALIZE_SERVICE = f"""{{"attributes": {{"SCPDURL": "{DATA_SERVICE_SCPDURL}", "controlURL": "{DATA_SERVICE_CONTROLURL}", "eventSubURL": "{DATA_SERVICE_EVENTSUBURL}", "serviceId": "{DATA_SERVICE_SERVICEID}", "serviceType": "{DATA_SERVICE_SERVICETYPE}"}}, "scpd": {JSON_RESULT_TEST_SERIALIZE_SCPD}}}"""
JSON_RESULT_TEST_SERIALIZE_DEVICE_BASIC = f"""{{"attributes": {{"UDN": "{DATA_DEVICE_UDN}", "UPC": null, "deviceType": "{DATA_DEVICE_DEVICETYPE}", "friendlyName": "{DATA_DEVICE_FRIENDLYNAME}", "manufacturer": "{DATA_DEVICE_MANUFACTURER}", "manufacturerURL": "{DATA_DEVICE_MANUFACTURERURL}", "modelDescription": "{DATA_DEVICE_MODELDESCRIPTION}", "modelName": "{DATA_DEVICE_MODELNAME}", "modelNumber": "{DATA_DEVICE_MODELNUMBER}", "modelURL": "{DATA_DEVICE_MODELURL}", "presentationURL": "{DATA_DEVICE_PRESENTATIONURL}"}}"""
JSON_RESULT_TEST_SERIALIZE_DEVICE_BASIC_SERVICE = f""", "services": [{JSON_RESULT_TEST_SERIALIZE_SERVICE}, {JSON_RESULT_TEST_SERIALIZE_SERVICE}]}}"""
JSON_RESULT_TEST_SERIALIZE_DEVICE = JSON_RESULT_TEST_SERIALIZE_DEVICE_BASIC + f""", "devices": [], "services": []}}"""  # noqa: F541
JSON_RESULT_TEST_SERIALIZE_DEVICE_WITH_SERVICES = JSON_RESULT_TEST_SERIALIZE_DEVICE_BASIC + ', "devices": []' + JSON_RESULT_TEST_SERIALIZE_DEVICE_BASIC_SERVICE

# subdevices don't provide services and further devices
JSON_RESULT_TEST_SERIALIZE_DEVICE_WITH_SERVICES_AND_SUBDEVICES = JSON_RESULT_TEST_SERIALIZE_DEVICE_BASIC + f""", "devices": [{JSON_RESULT_TEST_SERIALIZE_DEVICE}, {JSON_RESULT_TEST_SERIALIZE_DEVICE}]""" + JSON_RESULT_TEST_SERIALIZE_DEVICE_BASIC_SERVICE

JSON_RESULT_TEST_SERIALIZE_DESCRIPTION = f"""{{"device": {JSON_RESULT_TEST_SERIALIZE_DEVICE_WITH_SERVICES_AND_SUBDEVICES}, "specVersion": {JSON_RESULT_TEST_SERIALIZE_SPECVERSION}, "systemVersion": {JSON_RESULT_TEST_SERIALIZE_SYSTEMVERSION}}}"""
# JSON_RESULT_TEST_SERIALIZE_DESCRIPTION = f"""{{"systemVersion": {JSON_RESULT_TEST_SERIALIZE_SYSTEMVERSION}}}"""


def make_spec_version():
    spec_version = SpecVersion()
    spec_version.major = DATA_SPEC_MAJOR_VERSION
    spec_version.minor = DATA_SPEC_MINOR_VERSION
    return spec_version


def make_system_version():
    system_version = SystemVersion()
    system_version.HW = None
    system_version.Major = DATA_SYS_MAJOR
    system_version.Minor = DATA_SYS_MINOR
    system_version.Patch = DATA_SYS_PATCH
    system_version.Buildnumber = None
    system_version.Display = DATA_SYS_DISPLAY
    return system_version


def make_argument():
    argument = Argument()
    argument.name = DATA_ARGUMENT_NAME
    argument.direction = DATA_ARGUMENT_DIRECTION
    argument.relatedStateVariable = DATA_ARGUMENT_RELATEDSTATEVARIABLE
    return argument


def make_valuerange():
    valuerange = ValueRange()
    valuerange.minimum = DATA_VALUERANGE_MINIMUM
    valuerange.maximum = DATA_VALUERANGE_MAXIMUM
    valuerange.step = DATA_VALUERANGE_STEP
    return valuerange


def make_action():
    # an Action has a name and a list of arguments.
    # make an Action with two arguments.
    action = Action()
    action.name = DATA_ACTION_NAME
    action._arguments.append(make_argument())
    action._arguments.append(make_argument())
    return action


def make_statevariable():
    # a StateVariable has a name like 'WANAccessType'
    # a dataType like 'string' or 'ui2'
    # a defaultValue
    # a list of strings of allowed_values like 'DSL', 'Cable' etc.
    # an allowedValueRange
    # Not all Atrributes must be defined.
    statevariable = StateVariable()
    statevariable.name = DATA_STATEVARIABLE_NAME
    statevariable.dataType = DATA_STATEVARIABLE_DATATYPE
    statevariable.defaultValue = DATA_STATEVARIABLE_DEFAULTVALUE
    statevariable.allowed_values = DATA_STATEVARIABLE_ALLOWED_VALUES
    statevariable.allowedValueRange = make_valuerange()
    return statevariable


def make_scpd():
    # a service control point definition has a list of Actions,
    # a list of StateVariables and a SpecVersion.
    scpd = Scpd(root=[])
    scpd._actions = [make_action(), make_action()]
    scpd._state_variables = [make_statevariable(), make_statevariable()]
    scpd.specVersion = make_spec_version()
    return scpd


def make_service():
    service = Service()
    service._scpd = make_scpd()
    service.serviceType = DATA_SERVICE_SERVICETYPE
    service.serviceId = DATA_SERVICE_SERVICEID
    service.controlURL = DATA_SERVICE_CONTROLURL
    service.eventSubURL = DATA_SERVICE_EVENTSUBURL
    service.SCPDURL = DATA_SERVICE_SCPDURL
    return service


def make_device(with_services=True, with_subdevices=True):
    device = Device()
    device.deviceType = DATA_DEVICE_DEVICETYPE
    device.friendlyName = DATA_DEVICE_FRIENDLYNAME
    device.manufacturer = DATA_DEVICE_MANUFACTURER
    device.manufacturerURL = DATA_DEVICE_MANUFACTURERURL
    device.modelDescription = DATA_DEVICE_MODELDESCRIPTION
    device.modelName = DATA_DEVICE_MODELNAME
    device.modelNumber = DATA_DEVICE_MODELNUMBER
    device.modelURL = DATA_DEVICE_MODELURL
    device.UDN = DATA_DEVICE_UDN
    device.UPC = None
    device.presentationURL = DATA_DEVICE_PRESENTATIONURL
    if with_services:
        device._services = [make_service(), make_service()]
    if with_subdevices:
        device.devices = [
            make_device(with_services=False, with_subdevices=False),
            make_device(with_services=False, with_subdevices=False)
        ]
    return device


def make_description():
    description = Description(root=[])
    description.device = make_device()
    description.specVersion = make_spec_version()
    description.systemVersion = make_system_version()
    return description


class Check(Serializer):
    """
    Simple Subclass for Serializer for testing.
    """
    def __init__(self, one=1, two=2):
        self.one = one
        self.two = two


def test_serializer():
    """
    Test the Serializer superclass.
    """
    check = Check()
    data = check.serialize()
    one_three = Check(one=3)
    assert check != one_three
    one_three.deserialize(data)
    assert check == one_three


def test_serializer_02():
    """
    Test for same set of attributes.
    """
    small = Check()
    big = Check()
    assert small == big
    assert big == small
    big.three = 3
    assert small != big
    assert big != small


def test_serialize_exclude():
    """
    Test Serializer superclass with exclude argument
    """
    check = Check()
    data = check.serialize(exclude=['one'])
    one_three = Check(one=3)
    assert check != one_three
    one_three.deserialize(data)
    assert check != one_three
    one_three.one = 1
    assert check == one_three


def test_serialize_spec_version():
    """
    Serialize a SpecVersion instance to json.
    """
    spec_version = make_spec_version()
    result = spec_version.serialize()
    json_result = json.dumps(result)
    assert json_result == JSON_RESULT_TEST_SERIALIZE_SPECVERSION


def test_deserialize_spec_version():
    """
    Deserialize a SpecVersion instance from json.
    """
    data = json.loads(JSON_RESULT_TEST_SERIALIZE_SPECVERSION)
    sv = SpecVersion()
    sv.deserialize(data)
    assert sv == make_spec_version()


def test_serialize_system_version():
    """
    Serialize a SystemVersion instance to json.
    """
    system_version = make_system_version()
    result = system_version.serialize()
    json_result = json.dumps(result)
    assert json_result == JSON_RESULT_TEST_SERIALIZE_SYSTEMVERSION


def test_deserialize_system_version():
    """
    Deserialize a SystemVersion instance from json.
    """
    system_version = make_system_version()
    data = json.loads(JSON_RESULT_TEST_SERIALIZE_SYSTEMVERSION)
    sv = SystemVersion()
    assert sv != system_version
    sv.deserialize(data)
    assert sv == system_version


def test_serialize_argument():
    """Serialize an Argument to json."""
    argument = make_argument()
    result = argument.serialize()
    json_result = json.dumps(result)
    assert json_result == JSON_RESULT_TEST_SERIALIZE_ARGUMENT


def test_deserialize_argument():
    """Deserialize an Argument from json."""
    argument = make_argument()
    data = json.loads(JSON_RESULT_TEST_SERIALIZE_ARGUMENT)
    a = Argument()
    assert a != argument
    a.deserialize(data)
    assert a == argument


def test_serialize_valuerange():
    """Serialize a ValueRange to json."""
    valuerange = make_valuerange()
    result = json.dumps(valuerange.serialize())
    assert result == JSON_RESULT_TEST_SERIALIZE_VALUERANGE


def test_deserialize_valuerange():
    """Deserialize a ValueRange from json."""
    valuerange = make_valuerange()
    vr = ValueRange()
    assert vr != valuerange
    vr.deserialize(json.loads(JSON_RESULT_TEST_SERIALIZE_VALUERANGE))
    assert vr == valuerange


def test_serialize_action():
    """
    Serialize an Action with a name and two Argument instances in
    _arguments
    """
    action = make_action()
    result = json.dumps(action.serialize())
    assert result == JSON_RESULT_TEST_SERIALIZE_ACTION


def test_deserialize_action():
    """
    Deerialize an Action with a name and two Argument instances in
    _arguments from json.
    """
    action = make_action()
    ac = Action()
    ac.deserialize(json.loads(JSON_RESULT_TEST_SERIALIZE_ACTION))
    assert ac == action


def test_serialize_statevariable():
    statevariable = make_statevariable()
    result = json.dumps(statevariable.serialize())
    assert result == JSON_RESULT_TEST_SERIALIZE_STATEVARIABLE


def test_deserialize_statevariable():
    statevariable = make_statevariable()
    sv = StateVariable()
    assert sv != statevariable
    sv.deserialize(json.loads(JSON_RESULT_TEST_SERIALIZE_STATEVARIABLE))
    assert sv == statevariable


def test_serialize_scpd():
    scpd = make_scpd()
    result = json.dumps(scpd.serialize())
    assert result == JSON_RESULT_TEST_SERIALIZE_SCPD


def test_deserialize_scpd():
    scpd = make_scpd()
    s = Scpd(root=[])
    s.deserialize(json.loads(JSON_RESULT_TEST_SERIALIZE_SCPD))
    assert s == scpd


def test_serialize_service():
    """
    Serialize a Service Instance.
    """
    service = make_service()
    result = json.dumps(service.serialize())
    assert result == JSON_RESULT_TEST_SERIALIZE_SERVICE


def test_deserialize_service():
    """
    Deserialize a Service Instance.
    """
    service = make_service()
    data = json.loads(JSON_RESULT_TEST_SERIALIZE_SERVICE)
    s = Service()
    s.deserialize(data)
    assert s == service


def test_serialize_device():
    device = make_device(with_services=False, with_subdevices=False)
    result = json.dumps(device.serialize())
    assert result == JSON_RESULT_TEST_SERIALIZE_DEVICE


def test_deserialize_device():
    device = make_device(with_services=False, with_subdevices=False)
    d = Device()
    d.deserialize(json.loads(JSON_RESULT_TEST_SERIALIZE_DEVICE))
    assert d == device


def test_serialize_device_with_services():
    device = make_device(with_services=True, with_subdevices=False)
    result = json.dumps(device.serialize())
    assert result == JSON_RESULT_TEST_SERIALIZE_DEVICE_WITH_SERVICES


def test_deserialize_device_with_services():
    device = make_device(with_services=True, with_subdevices=False)
    d = Device()
    d.deserialize(json.loads(JSON_RESULT_TEST_SERIALIZE_DEVICE_WITH_SERVICES))
    assert d == device


def test_serialize_device_with_services_and_subdevices():
    device = make_device(with_services=True, with_subdevices=True)
    result = json.dumps(device.serialize())
    assert (
        result == JSON_RESULT_TEST_SERIALIZE_DEVICE_WITH_SERVICES_AND_SUBDEVICES
    )

def test_deserialize_device_with_services_and_subdevices():
    device = make_device(with_services=True, with_subdevices=True)
    d = Device()
    d.deserialize(
        json.loads(
            JSON_RESULT_TEST_SERIALIZE_DEVICE_WITH_SERVICES_AND_SUBDEVICES
        )
    )
    assert d == device


def test_serialize_description():
    description = make_description()
    result = json.dumps(description.serialize())
    assert result == JSON_RESULT_TEST_SERIALIZE_DESCRIPTION


def test_deserialize_description():
    description = make_description()
    data = json.loads(JSON_RESULT_TEST_SERIALIZE_DESCRIPTION)
    d = Description(root=[])
    d.deserialize(data)
    assert d == description
    d = Description.from_data(data)
    assert d == description



import io
import logging
import pytest
from ..core.logger import (
    fritzlogger,
    activate_local_debug_mode,
    reset,
)

stream = io.StringIO()
stream_handler = logging.StreamHandler(stream=stream)
formatter = logging.Formatter("%(levelname)s:%(message)s")
stream_handler.setFormatter(formatter)


def get_last_log_entry():
    lines = stream_handler.stream.getvalue().splitlines()
    if lines:
        return lines[-1]
    return ''


def expected_log_entry(message):
    # helper function for formatting
    return f"{message.upper()}:{message}"


def delete_log_entries():
    stream_handler.stream.close()
    stream_handler.stream = io.StringIO()


@pytest.mark.parametrize(
    "message, log_call, should_log", [
        ("debug", fritzlogger.debug, False),
        ("info", fritzlogger.info, True),
        ("warning", fritzlogger.warning, True),
        ("error", fritzlogger.error, True),
        ("critical", fritzlogger.critical, True),
    ])
def test_default_logging(message, log_call, should_log):
    fritzlogger.addHandler(stream_handler)
    delete_log_entries()
    log_call(message)
    last_message = get_last_log_entry()
    if should_log:
        expected = expected_log_entry(message)
    else:
        expected = ""
    assert last_message == expected


def test_local_debug_mode_no_propagate_no_handler():
    delete_log_entries()
    reset()
    activate_local_debug_mode()
    message = "debug"
    fritzlogger.debug(message)
    # no handler (NullHandler), no propagate:
    assert get_last_log_entry() == ""


def test_local_debug_mode_no_propagate_with_handler():
    delete_log_entries()
    reset()
    activate_local_debug_mode(handler=stream_handler)
    message = "debug"
    fritzlogger.debug(message)
    assert get_last_log_entry() == expected_log_entry(message)


def test_local_debug_mode_no_handler():
    delete_log_entries()
    reset()
    activate_local_debug_mode(propagate=True)
    message = "debug"
    fritzlogger.debug(message)
    # no handler, but propagate:
    assert get_last_log_entry() == ""


def test_local_debug_mode_no_handler_but_propagate_and_warning():
    # if emitted as warning, the message gets logged from the
    # parent-handlers (in this case root)
    # in this case the RootLogger needs a handler
    delete_log_entries()
    reset()
    activate_local_debug_mode(propagate=True)
    message = "warning"
    fritzlogger.warning(message)
    assert get_last_log_entry() == ""
    # provide a handler for root:
    logging.root.addHandler(stream_handler)
    fritzlogger.warning(message)
    assert get_last_log_entry() == expected_log_entry(message)
    logging.root.removeHandler(stream_handler)




import os
from pathlib import Path
from types import SimpleNamespace

import pytest

from ..core.devices import DeviceManager
from ..core.exceptions import FritzConnectionException
from ..core.fritzconnection import (
    FritzConnection,
    FRITZ_CACHE_DIR,
    FRITZ_CACHE_FORMAT_JSON,
    FRITZ_CACHE_FORMAT_PICKLE,
    FRITZ_CACHE_JSON_SUFFIX,
    FRITZ_CACHE_PICKLE_SUFFIX,
)


@pytest.mark.parametrize(
    "name, expected", [
        ('WANIPConn', 'WANIPConn1'),
        ('WANIPConn1', 'WANIPConn1'),
        ('WANIPConn:1', 'WANIPConn1'),
        ('WANIPConn2', 'WANIPConn2'),
        ('WANIPConn:2', 'WANIPConn2'),
        ('WANIPConn21', 'WANIPConn21'),
        ('WANIPConn:21', 'WANIPConn21'),
    ]
)
def test_normalize_name(name, expected):
    result = FritzConnection.normalize_name(name)
    assert result == expected


@pytest.mark.parametrize(
    "url, use_tls, expected", [
        ('192.168.178.1', False, 'http://192.168.178.1'),
        ('192.168.178.1', True, 'https://192.168.178.1'),
        ('http://192.168.178.1', False, 'http://192.168.178.1'),
        ('http://192.168.178.1', True, 'https://192.168.178.1'),
        ('https://192.168.178.1', False, 'http://192.168.178.1'),
        ('https://192.168.178.1', True, 'https://192.168.178.1'),
    ]
)
def test_set_protocol(url, use_tls, expected):
    result = FritzConnection.set_protocol(url, use_tls)
    assert result == expected


@pytest.mark.parametrize(
    "address, port, cache_directory, cache_format, expected_result", [
        ('192.168.178.1', 49000, 'some_path', FRITZ_CACHE_FORMAT_PICKLE, Path("some_path") / f'192_168_178_1_cache.{FRITZ_CACHE_PICKLE_SUFFIX}'),
        ('192.168.178.1', 49000, '', FRITZ_CACHE_FORMAT_PICKLE, Path().home() / FRITZ_CACHE_DIR / f'192_168_178_1_cache.{FRITZ_CACHE_PICKLE_SUFFIX}'),
        ('192.168.178.1', 49000, None, FRITZ_CACHE_FORMAT_PICKLE, Path().home() / FRITZ_CACHE_DIR / f'192_168_178_1_cache.{FRITZ_CACHE_PICKLE_SUFFIX}'),
        ('192.168.178.6', 49000, None, FRITZ_CACHE_FORMAT_PICKLE, Path().home() / FRITZ_CACHE_DIR / f'192_168_178_6_cache.{FRITZ_CACHE_PICKLE_SUFFIX}'),
        ('192.168.1.15', 49000, None, FRITZ_CACHE_FORMAT_PICKLE, Path().home() / FRITZ_CACHE_DIR / f'192_168_1_15_cache.{FRITZ_CACHE_PICKLE_SUFFIX}'),
        ('http://192.168.1.15', 49000, None, FRITZ_CACHE_FORMAT_PICKLE, Path().home() / FRITZ_CACHE_DIR / f'192_168_1_15_cache.{FRITZ_CACHE_PICKLE_SUFFIX}'),
        ('https://192.168.1.15', 49000, None, FRITZ_CACHE_FORMAT_PICKLE, Path().home() / FRITZ_CACHE_DIR / f'192_168_1_15_cache.{FRITZ_CACHE_PICKLE_SUFFIX}'),
        ('192.168.178.1', 49000, 'some_path', FRITZ_CACHE_FORMAT_JSON, Path("some_path") / f'192_168_178_1_cache.{FRITZ_CACHE_JSON_SUFFIX}'),
        ('192.168.178.1', 49000, '', FRITZ_CACHE_FORMAT_JSON, Path().home() / FRITZ_CACHE_DIR / f'192_168_178_1_cache.{FRITZ_CACHE_JSON_SUFFIX}'),
        ('192.168.178.1', 49000, None, FRITZ_CACHE_FORMAT_JSON, Path().home() / FRITZ_CACHE_DIR / f'192_168_178_1_cache.{FRITZ_CACHE_JSON_SUFFIX}'),
        ('192.168.178.6', 49000, None, FRITZ_CACHE_FORMAT_JSON, Path().home() / FRITZ_CACHE_DIR / f'192_168_178_6_cache.{FRITZ_CACHE_JSON_SUFFIX}'),
        ('192.168.1.15', 49000, None, FRITZ_CACHE_FORMAT_JSON, Path().home() / FRITZ_CACHE_DIR / f'192_168_1_15_cache.{FRITZ_CACHE_JSON_SUFFIX}'),
        ('http://192.168.1.15', 49000, None, FRITZ_CACHE_FORMAT_JSON, Path().home() / FRITZ_CACHE_DIR / f'192_168_1_15_cache.{FRITZ_CACHE_JSON_SUFFIX}'),
        ('https://192.168.1.15', 49000, None, FRITZ_CACHE_FORMAT_JSON, Path().home() / FRITZ_CACHE_DIR / f'192_168_1_15_cache.{FRITZ_CACHE_JSON_SUFFIX}'),
    ]
)
def test_get_cache_path(address, port, cache_directory, cache_format, expected_result):
    # port parameter intentionally unused
    mock = SimpleNamespace(address=address, port=port)
    result = FritzConnection._get_cache_path(mock, cache_directory, cache_format)
    assert result == expected_result


def test_invalide_cache_format():
    mock = SimpleNamespace(address='192.168.178.1', port=49000)
    with pytest.raises(FritzConnectionException):
        _ = FritzConnection._get_cache_path(mock, None, 'weird')


@pytest.mark.parametrize(
    "file_name, cache_format", [
        ("description.json", FRITZ_CACHE_FORMAT_JSON),
        ("description.pcl", FRITZ_CACHE_FORMAT_PICKLE),
    ]
)
def test_load_api_from_cache(file_name, cache_format, datadir):
    """
    Load the api from a cache file and check whether all
    known services are available.
    """
    os.chdir(datadir)
    mock = SimpleNamespace(device_manager=DeviceManager())
    FritzConnection._load_api_from_cache(mock, file_name, cache_format)
    with open("description_service_names.txt") as fobj:
        service_names = list(map(str.strip, fobj))
    known_services = mock.device_manager.services.keys()
    # check for same number of services:
    assert len(known_services) == len(service_names)
    # check for same service names:
    for name in service_names:
        assert name in known_services


@pytest.mark.parametrize(
    "file_name, cache_format", [
        ("description.json", FRITZ_CACHE_FORMAT_JSON),
        ("description.pcl", FRITZ_CACHE_FORMAT_PICKLE),
    ]
)
def test_write_api_to_cache(file_name, cache_format, datadir):
    """
    Tests writing the api data.

    For writing there must be some data first. If the test
    `test_load_api_from_cache` succeeds, the data can safely loaded to a
    FritzConnection mock object. The data are then exported to a file
    and loaded again to another FritzConnection mock object. Then the
    Decription objects are compared.
    """
    os.chdir(datadir)
    # 1. load the api data
    mock = SimpleNamespace(device_manager=DeviceManager())
    FritzConnection._load_api_from_cache(mock, file_name, cache_format)
    # 2. save the api data
    cache_filename = f"x_tmp_cache_{file_name}"
    cache_file = Path(cache_filename)

# TODO: (potentially)
# code for Python >= 3.8:
# cache_file.unlink(missing_ok=True)  # in case of artefacts
# code for Python < 3.8:
    try:
        cache_file.unlink()
    except FileNotFoundError:
        pass  # ignore

    FritzConnection._write_api_to_cache(mock, cache_filename, cache_format)
    assert cache_file.exists() is True
    # 3. read the api data to another mock
    other_mock = SimpleNamespace(device_manager=DeviceManager())
    FritzConnection._load_api_from_cache(other_mock, cache_filename, cache_format)
    # 4. the Descriptions are serializable and therefore comparable objects:
    for desc1, desc2 in zip(
        mock.device_manager.descriptions, other_mock.device_manager.descriptions
    ):
        assert desc1 == desc2
    # clean up
    cache_file.unlink()
    assert cache_file.exists() is False





import datetime
import types
from xml.etree import ElementTree as etree

import pytest

from ..core.exceptions import (
    ActionError,
    FritzAuthorizationError,
    FritzConnectionException,
    FritzActionError,
    FritzArgumentError,
    FritzActionFailedError,
    FritzArgumentValueError,
    FritzOutOfMemoryError,
    FritzSecurityError,
    FritzArrayIndexError,
    FritzLookUpError,
    FritzArgumentStringToShortError,
    FritzArgumentStringToLongError,
    FritzArgumentCharacterError,
    FritzInternalError,
)

from ..core.soaper import (
    boolean_convert,
    encode_boolean,
    get_argument_value,
    get_converted_value,
    get_html_safe_value,
    is_html_response,
    raise_fritzconnection_error,
    redact_response,
    remove_html_tags,
)


content_template = """
<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
  <s:Body>
    <s:Fault>
      <faultcode>s:Client</faultcode>
      <faultstring>UPnPError</faultstring>
      <detail>
        <UPnPError xmlns="urn:schemas-upnp-org:control-1-0">
          <errorCode>{error_code}</errorCode>
          <errorDescription>Invalid Action</errorDescription>
        </UPnPError>
      </detail>
    </s:Fault>
  </s:Body>
</s:Envelope>
"""


@pytest.mark.parametrize(
    "error_code, status_code, exception", [
        ('401', 500, FritzActionError),
        ('402', 500, FritzArgumentError),
        ('501', 500, FritzActionFailedError),
        ('600', 500, FritzArgumentValueError),
        ('603', 500, FritzOutOfMemoryError),
        ('606', 500, FritzSecurityError),
        ('713', 500, FritzArrayIndexError),
        ('714', 500, FritzLookUpError),
        ('801', 500, FritzArgumentStringToShortError),
        ('802', 500, FritzArgumentStringToLongError),
        ('803', 500, FritzArgumentCharacterError),
        ('820', 500, FritzInternalError),

        ('713', 500, IndexError),
        ('714', 500, KeyError),

        ('401', 500, ActionError),

        (None, 401, FritzAuthorizationError),
        (None, 500, FritzConnectionException),

    ]
)
def test_raise_fritzconnection_error(error_code, status_code, exception):
    """check for exception raising depending on the error_code"""
    if error_code:
        content = content_template.format(error_code=error_code)
    else:
        content = '<html>some content</html>'
    response = types.SimpleNamespace()
    response.text = content
    response.content = content.encode()
    response.status_code = status_code
    pytest.raises(exception, raise_fritzconnection_error, response)


def test_raise_fritzauthorization_error():
    """check for exception raising depending on the html status code."""
    response = types.SimpleNamespace()
    response.content = b'<HTML><HEAD><TITLE>401 Unauthorized (ERR_NONE)</TITLE></HEAD><BODY><H1>401 Unauthorized</H1><BR>ERR_NONE<HR><B>Webserver</B> Sat, 01 Oct 2022 09:46:22 GMT</BODY></HTML>'
    response.text = '<HTML><HEAD><TITLE>401 Unauthorized (ERR_NONE)</TITLE></HEAD><BODY><H1>401 Unauthorized</H1><BR>ERR_NONE<HR><B>Webserver</B> Sat, 01 Oct 2022 09:46:22 GMT</BODY></HTML>'
    response.status_code = 401
    pytest.raises(FritzAuthorizationError, raise_fritzconnection_error, response)

    # simulate a http/500
    response.content = b'<HTML><HEAD><TITLE>500 internal error</TITLE></HEAD><BODY><H1>500 internal error</H1><BR>ERR_NONE<HR><B>Webserver</B> Sat, 01 Oct 2022 09:46:22 GMT</BODY></HTML>'
    response.text = '<HTML><HEAD><TITLE>500 internal error</TITLE></HEAD><BODY><H1>500 internal error</H1><BR>ERR_NONE<HR><B>Webserver</B> Sat, 01 Oct 2022 09:46:22 GMT</BODY></HTML>'
    response.status_code = 500
    pytest.raises(FritzConnectionException, raise_fritzconnection_error, response)


@pytest.mark.parametrize(
    "value, expected_result", [
        ('0', False),
        ('1', True),
    ]
)
def test_boolean_convert(value, expected_result):
    result = boolean_convert(value)
    assert result == expected_result


@pytest.mark.parametrize(
    "value", ['2', 'x', '3.1']
)
def test_boolean_convert_fails(value):
    with pytest.raises(ValueError):
        boolean_convert(value)


@pytest.mark.parametrize(
    "text, expected_result", [
        ('<html>something</html>', True),
        ('<?xml version= ...', False),
    ]
)
def test_is_html_response(text, expected_result):
    result = is_html_response(text)
    assert result == expected_result


@pytest.mark.parametrize(
    "text, expected_result", [
        ("funky", "funky"),
        ("funky funky", "funky funky"),
        ("funky  funky", "funky funky"),
        ("<bold>here</bold>", "here"),
        ("<bold>ham <i>spam</i><hr>and eggs </bold>", "ham spam and eggs"),
        (
            "<html><head>failure</head><body>something <b>very</b><hr /><i>strange</i></body></html>",
            "failure something very strange",
        )
    ]
)
def test_remove_html_tags(text, expected_result):
    result = remove_html_tags(text)
    assert result == expected_result


long_error = """
<s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
  <s:Body>
    <s:Fault>
      <faultcode> s:Client </faultcode>
      <faultstring>
UPnPError </faultstring>
      <detail>
        <UPnPError xmlns="urn:schemas-upnp-org:control-1-0">
          <errorCode> 401 </errorCode>
          <errorDescription> Invalid Action </errorDescription>
        </UPnPError>
      </detail>
    </s:Fault>
  </s:Body>
</s:Envelope>
"""


def test_long_error_message():
    response = types.SimpleNamespace()
    response.text = long_error
    response.content = long_error.encode()
    with pytest.raises(ActionError) as exc:
        raise_fritzconnection_error(response)

    assert exc.value.args[0] == "\n".join(
        ["UPnPError: ",
         "errorCode: 401",
         "errorDescription: Invalid Action",
         ]
    )


@pytest.mark.parametrize(
    "value, expected_type", [
        ("text", str),
        (0, int),
        (1, int),
        (None, int),
        (False, int),
        (True, int),
    ]
)
def test_encode_boolean(value, expected_type):
    result = encode_boolean(value)
    assert isinstance(result, expected_type)


@pytest.mark.parametrize(
    "value, expected_results", [
        (True, 1),
        (False, 0),
        (None, 0),
        (3.141, 3.141),
        ("hello test", "hello test"),
        ("2021-07-17T12:00:00", "2021-07-17T12:00:00"),  # redundant, but ISO ;)
        ("ham, spam & eggs", "ham, spam &amp; eggs"),
        ("5 > 3", "5 &gt; 3"),
        ("3 < 5", "3 &lt; 5"),
        ('say "hello"', "say &quot;hello&quot;"),
        ("let's test again", ["let&apos; test again", "let&#x27;s test again"])
    ]
)
def test_get_html_safe_value(value, expected_results):
    if not isinstance(expected_results, list):
        expected_results = [expected_results]
    result = get_html_safe_value(value)
    assert result in expected_results


@pytest.mark.parametrize(
    "value, not_expected_type", [
        (False, bool),  # should be int after encoding, not bool
        (True, bool),
    ]
)
def test_encode_boolean2(value, not_expected_type):
    result = encode_boolean(value)
    assert not isinstance(result, not_expected_type)


soap_root = etree.fromstring("""<?xml version="1.0"?>
<data>
    <container>
        <year>2010</year>
        <msg>message text</msg>
        <number>3.141</number>
        <ip></ip>
    </container>
</data>""")


@pytest.mark.parametrize(
    "argument_name, expected_value", [
        ('year', '2010'),
        ('msg', 'message text'),
        ('number', '3.141'),
        ('ip', ''),
    ]
)
def test_get_argument_value(argument_name, expected_value):
     value = get_argument_value(soap_root, argument_name)
     assert value == expected_value


@pytest.mark.parametrize(
    "data_type, value, expected_value", [
        ('datetime', '2020-02-02T10:10:10', datetime.datetime(2020, 2, 2, 10, 10, 10)),
        ('boolean', '1', True),
        ('boolean', '0', False),
        ('uuid', 'uuid:123', '123'),
        ('uuid', '123', '123'),
        ('i4', '42', 42),
        ('ui1', '42', 42),
        ('ui2', '42', 42),
        ('ui4', '42', 42),
    ]
)
def test_get_converted_value(data_type, value, expected_value):
    result = get_converted_value(data_type, value)
    assert result == expected_value


@pytest.mark.parametrize(
    "data_type, value", [
        ('datetime', '2010.02.02-10:10:10'),  # not ISO 8601
        ('boolean', ''),  # neither '1' nor '0'
    ]
)
def test_get_converted_value_fails(data_type, value):
    with pytest.raises(ValueError):
        get_converted_value(data_type, value)

def test_redact_debug_log_phone_numbers():
    response = """
    <?xml version="1.0"?>
    <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
    <s:Body>
    <u:GetInfoResponse xmlns:u="urn:dslforum-org:service:DeviceInfo:1">
    <NewManufacturerName>AVM</NewManufacturerName>
    <NewManufacturerOUI>00040E</NewManufacturerOUI>
    <NewModelName>FRITZ!Box 7530 AX</NewModelName>
    <NewDescription>FRITZ!Box 7530 AX Release 256.08.00</NewDescription>
    <NewProductClass>FRITZ!Box</NewProductClass>
    <NewSerialNumber>aabbccddeeff</NewSerialNumber>
    <NewSoftwareVersion>256.08.00</NewSoftwareVersion>
    <NewHardwareVersion>FRITZ!Box 7530 AX</NewHardwareVersion>
    <NewSpecVersion>1.0</NewSpecVersion>
    <NewProvisioningCode></NewProvisioningCode>
    <NewUpTime>86446</NewUpTime>
    <NewDeviceLog>
    23.11.24 12:28:10 Anmeldung der Internetrufnummer 491234567890 war nicht erfolgreich. Ursache: DNS-Fehler
    23.11.24 12:28:10 Anmeldung der Internetrufnummer 491234567891 war nicht erfolgreich. Ursache: DNS-Fehler
    23.11.24 12:28:10 Anmeldung der Internetrufnummer 491234567892 war nicht erfolgreich. Ursache: DNS-Fehler
    23.11.24 12:28:10 Anmeldung der Internetrufnummer 491234567893 war nicht erfolgreich. Ursache: DNS-Fehler
    </u:GetInfoResponse>
    </s:Body>
    </s:Envelope>
    """

    result = redact_response(False, response)
    assert result == response

    result = redact_response(True, response)
    assert result == """
    <?xml version="1.0"?>
    <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
    <s:Body>
    <u:GetInfoResponse xmlns:u="urn:dslforum-org:service:DeviceInfo:1">
    <NewManufacturerName>AVM</NewManufacturerName>
    <NewManufacturerOUI>00040E</NewManufacturerOUI>
    <NewModelName>FRITZ!Box 7530 AX</NewModelName>
    <NewDescription>FRITZ!Box 7530 AX Release 256.08.00</NewDescription>
    <NewProductClass>FRITZ!Box</NewProductClass>
    <NewSerialNumber>aabbccddeeff</NewSerialNumber>
    <NewSoftwareVersion>256.08.00</NewSoftwareVersion>
    <NewHardwareVersion>FRITZ!Box 7530 AX</NewHardwareVersion>
    <NewSpecVersion>1.0</NewSpecVersion>
    <NewProvisioningCode></NewProvisioningCode>
    <NewUpTime>86446</NewUpTime>
    <NewDeviceLog>
    23.11.24 12:28:10 Anmeldung der Internetrufnummer ****** war nicht erfolgreich. Ursache: DNS-Fehler
    23.11.24 12:28:10 Anmeldung der Internetrufnummer ****** war nicht erfolgreich. Ursache: DNS-Fehler
    23.11.24 12:28:10 Anmeldung der Internetrufnummer ****** war nicht erfolgreich. Ursache: DNS-Fehler
    23.11.24 12:28:10 Anmeldung der Internetrufnummer ****** war nicht erfolgreich. Ursache: DNS-Fehler
    </u:GetInfoResponse>
    </s:Body>
    </s:Envelope>
    """

def test_redact_debug_log_external_ip_addresses():
    response = """
    <?xml version="1.0" encoding="utf-8"?>
    <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
    <s:Body>
    <u:GetExternalIPAddressResponse xmlns:u="urn:schemas-upnp-org:service:WANIPConnection:1">
    <NewExternalIPAddress>12.34.56.78</NewExternalIPAddress>
    </u:GetExternalIPAddressResponse>
    </s:Body>
    </s:Envelope>
    <?xml version="1.0" encoding="utf-8"?>
    <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
    <s:Body>
    <u:X_AVM_DE_GetExternalIPv6AddressResponse xmlns:u="urn:schemas-upnp-org:service:WANIPConnection:1">
    <NewExternalIPv6Address>0011:2233:4455:6677::0abcd</NewExternalIPv6Address>
    <NewPrefixLength>64</NewPrefixLength>
    <NewValidLifetime>0</NewValidLifetime>
    <NewPreferedLifetime>0</NewPreferedLifetime>
    </u:X_AVM_DE_GetExternalIPv6AddressResponse>
    </s:Body>
    </s:Envelope>
    """

    result = redact_response(False, response)
    assert result == response

    result = redact_response(True, response)
    assert result == """
    <?xml version="1.0" encoding="utf-8"?>
    <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
    <s:Body>
    <u:GetExternalIPAddressResponse xmlns:u="urn:schemas-upnp-org:service:WANIPConnection:1">
    <NewExternalIPAddress>******</NewExternalIPAddress>
    </u:GetExternalIPAddressResponse>
    </s:Body>
    </s:Envelope>
    <?xml version="1.0" encoding="utf-8"?>
    <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
    <s:Body>
    <u:X_AVM_DE_GetExternalIPv6AddressResponse xmlns:u="urn:schemas-upnp-org:service:WANIPConnection:1">
    <NewExternalIPv6Address>******</NewExternalIPv6Address>
    <NewPrefixLength>64</NewPrefixLength>
    <NewValidLifetime>0</NewValidLifetime>
    <NewPreferedLifetime>0</NewPreferedLifetime>
    </u:X_AVM_DE_GetExternalIPv6AddressResponse>
    </s:Body>
    </s:Envelope>
    """

def test_redact_debug_log_wifi_passwords():
    response = """
    <?xml version="1.0"?>
    <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
    <s:Body>
    <u:GetSecurityKeysResponse xmlns:u="urn:dslforum-org:service:WLANConfiguration:3">
    <NewWEPKey0>0123456789</NewWEPKey0>
    <NewWEPKey1></NewWEPKey1>
    <NewWEPKey2>01234 6789</NewWEPKey2>
    <NewWEPKey3></NewWEPKey3>
    <NewPreSharedKey>0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF0123456789ABCDEF</NewPreSharedKey>
    <NewKeyPassphrase>MY_GREAT_WIFI_PSK</NewKeyPassphrase>
    </u:GetSecurityKeysResponse>
    </s:Body>
    </s:Envelope>
    """

    result = redact_response(False, response)
    assert result == response

    result = redact_response(True, response)
    assert result == """
    <?xml version="1.0"?>
    <s:Envelope xmlns:s="http://schemas.xmlsoap.org/soap/envelope/" s:encodingStyle="http://schemas.xmlsoap.org/soap/encoding/">
    <s:Body>
    <u:GetSecurityKeysResponse xmlns:u="urn:dslforum-org:service:WLANConfiguration:3">
    <NewWEPKey0>******</NewWEPKey0>
    <NewWEPKey1>******</NewWEPKey1>
    <NewWEPKey2>******</NewWEPKey2>
    <NewWEPKey3>******</NewWEPKey3>
    <NewPreSharedKey>******</NewPreSharedKey>
    <NewKeyPassphrase>******</NewKeyPassphrase>
    </u:GetSecurityKeysResponse>
    </s:Body>
    </s:Envelope>
    """



"""
tests for qr-code creation
"""

import io
import os
import tempfile

import pytest

try:
    import cv2
    import segno.helpers
except ImportError:
    OPENCV_NOT_AVAILABLE = True
else:
    OPENCV_NOT_AVAILABLE = False

from fritzconnection.lib.fritzwlan import _get_wifi_qr_code as get_wifi_qr_code


@pytest.mark.skipif(OPENCV_NOT_AVAILABLE, reason="requires opencv")
def test_tools():
    """
    basic test whether opencv is able to read the output provided by segno.
    The tempfile is needed because opencv can not read from a file-like
    object but needs a path to the image-file.
    """
    ssid = 'xenon'
    password = 'the_strange_one'
    security = 'WPA'
    kind = 'png'
    # create qr-code and store it in a persistent temporary file
    stream = tempfile.NamedTemporaryFile(mode='wb', delete=False, suffix=f'.{kind}')
    fname = stream.name
    qr_code = segno.helpers.make_wifi(
        ssid=ssid, password=password, security=security, hidden=False
    )
    qr_code.save(out=stream, kind=kind, scale=4)
    stream.close()
    assert os.path.exists(fname) is True

    # test qr-code reading
    expected_result = f'WIFI:T:{security};S:{ssid};P:{password};;'
    img = cv2.imread(fname)
    detector = cv2.QRCodeDetector()
    result, _, _ = detector.detectAndDecode(img)
    assert result == expected_result

    # remove the tempfile
    os.unlink(fname)
    assert os.path.exists(fname) is False


def write_stream_to_tempfile(stream, kind='png'):
    """
    Helper function to store a segno stream in a persistent tempfile.
    Takes the stream (BytesIO) and returns the filepath (as string).
    """
    fobj = tempfile.NamedTemporaryFile(mode='wb', delete=False, suffix=f'.{kind}')
    fname = fobj.name
    fobj.write(stream.read())
    fobj.close()
    return fname


def get_content_from_qr_file(fname):
    """Helper function for qr-code reading."""
    img = cv2.imread(fname)
    detector = cv2.QRCodeDetector()
    result, _, _ = detector.detectAndDecode(img)
    return result


@pytest.mark.skipif(OPENCV_NOT_AVAILABLE, reason="requires opencv")
def test_helper_functions():
    ssid = 'xenon'
    password = 'another_strange_one'
    security = 'WPA'
    kind = 'png'
    stream = io.BytesIO()
    qr_code = segno.helpers.make_wifi(
        ssid=ssid, password=password, security=security, hidden=False
    )
    qr_code.save(out=stream, kind=kind, scale=4)
    stream.seek(0)
    # delegate stream to helper:
    fname = write_stream_to_tempfile(stream)
    # file must exist:
    assert os.path.exists(fname) is True
    # read qr-code:
    result = get_content_from_qr_file(fname)
    # check and clean up:
    expected_result = f'WIFI:T:{security};S:{ssid};P:{password};;'
    assert result == expected_result
    os.unlink(fname)
    assert os.path.exists(fname) is False


class WLANConfigMock:
    """
    Mocking class to provide the result of a
    WLANConfiguration.get_info() call, provide a get_password() method
    and a ssid attribute. All returned value must be part of the
    mock_data dictionary. The values in the dictionary are all strings.
    """

    def __init__(self, mock_data):
        self.mock_data = mock_data

    def get_info(self):
        return self.mock_data

    def get_password(self):
        # original from the action "GetSecurityKeys"
        return self.mock_data["NewKeyPassphrase"]

    @property
    def ssid(self):
        return self.mock_data["NewSSID"]


@pytest.mark.skipif(OPENCV_NOT_AVAILABLE, reason="requires opencv")
@pytest.mark.parametrize(
    "current_beacontype, security, expected_security", [
        ('11i', None, 'WPA'),
        ('11i', '', 'WPA'),
        ('11i', 'WPA3', 'WPA3'),
        ('WPAand11i', None, 'WPA'),
        ('11iandWPA3', None, 'WPA'),
        ('Basic', None, 'WEP'),
        ('None', None, 'nopass'),
        ('OWETrans', None, 'nopass'),
    ]
)
def test_get_wifi_qr_code(current_beacontype, security, expected_security):
    """
    Check for correct qr-code creation depending on the WLAN-settings.
    """
    ssid = "the_wlan_name"
    password = "the_password"
    expected_result = f"WIFI:T:{expected_security};S:{ssid};P:{password};;"
    mock_data = {
        'NewBeaconType': current_beacontype,
        'NewX_AVM-DE_PossibleBeaconTypes': 'None,11i,WPAand11i,11iandWPA3',
        'NewSSID': ssid,
        'NewKeyPassphrase': password,
    }
    instance = WLANConfigMock(mock_data)
    stream = get_wifi_qr_code(instance, kind="png", security=security)
    fname = write_stream_to_tempfile(stream)
    result = get_content_from_qr_file(fname)
    os.unlink(fname)  # do this asap
    assert result == expected_result



from pathlib import Path
import pytest


@pytest.fixture
def datadir():
    """
    Simple fixture for getting data files
    """
    here = Path(__file__).parent / "xml"
    return here


def pytest_configure(config):
    config.addinivalue_line("markers", "routertest: mark test accessing the router")




from ..lib.fritzhomeauto import HomeAutomationDevice


# take some real data from a switch
DEVICE_INFORMATION = {
    'NewAIN': '11657 0240192',
    'NewDeviceId': 16,
    'NewFunctionBitMask': 35712,
    'NewFirmwareVersion': '04.25',
    'NewManufacturer': 'AVM',
    'NewProductName': 'FRITZ!DECT 210',
    'NewDeviceName': 'FRITZ!DECT 210 #1',
    'NewPresent': 'CONNECTED',
    'NewMultimeterIsEnabled': 'ENABLED',
    'NewMultimeterIsValid': 'VALID',
    'NewMultimeterPower': 0,
    'NewMultimeterEnergy': 16659,
    'NewTemperatureIsEnabled': 'ENABLED',
    'NewTemperatureIsValid': 'VALID',
    'NewTemperatureCelsius': 195,
    'NewTemperatureOffset': 0,
    'NewSwitchIsEnabled': 'ENABLED',
    'NewSwitchIsValid': 'VALID',
    'NewSwitchState': 'OFF',
    'NewSwitchMode': 'AUTO',
    'NewSwitchLock': False,
    'NewHkrIsEnabled': 'DISABLED',
    'NewHkrIsValid': 'INVALID',
    'NewHkrIsTemperature': 0,
    'NewHkrSetVentilStatus': 'CLOSED',
    'NewHkrSetTemperature': 0,
    'NewHkrReduceVentilStatus': 'CLOSED',
    'NewHkrReduceTemperature': 0,
    'NewHkrComfortVentilStatus': 'CLOSED',
    'NewHkrComfortTemperature': 0
}

BASICDEVICESTATS_RESPONSE_FILENAME = "basicdevicestats_response.txt"


def test_homeautomation_device_properties():
    """
    Test whether a switch provides a property of a switch, but not other
    properties. A switch as 'FRITZ!DECT 210' is

        - an energy sensor
        - a temperature sensor
        - pluggable
        - and has a microphone

    """
    dp = HomeAutomationDevice(None, DEVICE_INFORMATION)
    assert dp.is_han_fun_unit is False
    assert dp.is_bulb is False
    assert dp.is_alarm_sensor is False
    assert dp.is_avm_button is False
    assert dp.is_radiator_control is False
    assert dp.is_energy_sensor is True
    assert dp.is_temperature_sensor is True
    assert dp.is_pluggable is True
    assert dp.is_avm_dect_repeater is False
    assert dp.is_microphone is True
    assert dp.is_switchable is True
    assert dp.is_adjustable is False
    assert dp.is_color_bulb is False
    assert dp.is_blind is False
    assert dp.is_humidity_sensor is False


def test_homeautomation_device_properties_with_ain():
    """
    Provides an additional ain to the one already in DEVICE_INFORMATION.
    The one in DEVICE_INFORMATION should be preserved.
    """
    original_ain = DEVICE_INFORMATION['NewAIN']
    other_ain = '00000 1111111'
    dp = HomeAutomationDevice(None, DEVICE_INFORMATION , identifier=other_ain)
    assert dp.identifier == original_ain


def test_homeautomation_device_properties_without_ain():
    """
    Provides an ain because the entry in DEVICE_INFORMATION is missing.
    """
    # don't modify the constant DEVICE_INFORMATION
    device_information = {
        k: v for k, v in DEVICE_INFORMATION.items() if k != "NewAIN"
    }
    other_ain = '00000 1111111'
    dp = HomeAutomationDevice(None, device_information , identifier=other_ain)
    assert dp.identifier == other_ain


def test_extract_basicdevicestats_response(datadir):
    """
    Inject some data from a real device with an energy sensor.
    """
    expected_counts = {
        "temperature": 96,
        "voltage": 360,
        "power": 360,
        "energy": 12
    }
    file = datadir / BASICDEVICESTATS_RESPONSE_FILENAME
    text = file.read_text(encoding="utf-8")
    response = {
        "content-type": "text/xml",
        "encoding": "charset=utf-8",
        "content": text.strip()
    }
    result = HomeAutomationDevice.extract_basicdevicestats_response(response)
    for key in ("temperature", "voltage", "power", "energy"):
        assert key in result
        count = result[key]['count']
        assert count == expected_counts[key]
        assert count == len(result[key]['data'])



import json
import os

import pytest

from ..core.devices import DeviceManager

@pytest.mark.parametrize("filename", [
    "igddesc.xml",
    "tr64desc.xml",
]
                         )
def test_DeviceManager_load_descriptions(device_manager, datadir, filename):
    """test loading of two description files."""
    os.chdir(datadir)
    device_manager.add_description(filename)
    assert device_manager.modelname == 'FRITZ!Box 7590'


@pytest.fixture()
def device_manager(datadir):
    os.chdir(datadir)
    manager = DeviceManager()
    manager.add_description('igddesc.xml')
    manager.add_description('tr64desc.xml')
    return manager


@pytest.mark.parametrize(
    "name", ['WANIPConn1', 'DeviceConfig1', 'WLANConfiguration1']
)
def test_DeviceManager_get_service(name, device_manager):
    """
    test services dict of the DeviceManager. Should find nested services
    (from sub devices) and from different files.
    """
    device_manager.scan()
    # should not raise a KeyError
    service = device_manager.services[name]
    assert service.name == name


@pytest.fixture()
def device_manager_from_data(datadir):
    device_manager = DeviceManager()
    os.chdir(datadir)
    with open("description.json") as fobj:
        data = json.load(fobj)
    device_manager.deserialize(data)
    return device_manager


def test_DeviceManager_load_json_check_descriptions(device_manager_from_data):
    """
    test build up of a DeviceManager from json data.
    """
    dm = device_manager_from_data
    assert len(dm.descriptions) == 2  # based on the test-data


def test_DeviceManager_load_json_check_services_01(device_manager_from_data):
    """
    check for loaded services after calling .scan()
    """
    dm = device_manager_from_data
    assert dm.services == {}  # still empty
    # run scan() should fill the services:
    dm.scan()
    assert dm.services != {}


def test_DeviceManager_load_json_check_services_02(device_manager_from_data):
    """
    check for matching services after calling .scan()
    """
    dm = device_manager_from_data
    dm.scan()
    # "description_service_names.txt" corresponds to "description.json"
    with open("description_service_names.txt") as fobj:
        service_names = list(map(str.strip, fobj))
    known_services = dm.services.keys()
    # check for same number of services:
    assert len(known_services) == len(service_names)
    # check for same service names:
    for name in service_names:
        assert name in known_services


def test_DeviceManager_load_json_check_actions(device_manager_from_data):
    """
    Simple test whether actions are available from the services (which
    in turn get the actions from the _scpd attribute).
    """
    dm = device_manager_from_data
    dm.scan()
    for service in dm.services.values():
        assert service.actions is not None



import nox


PYTHON_TEST_VERSIONS = ("3.7", "3.8", "3.9", "3.10", "3.11", "3.12", "3.13", "3.14")
PYTHON_DEVELOPMENT_VERSION = "3.11"


@nox.session(python=PYTHON_TEST_VERSIONS)
def test(session):
    session.install("-e", ".")
    session.install("requests", "pytest")
    session.run("pytest", "-m", "not routertest")


@nox.session(python=PYTHON_TEST_VERSIONS)
def test_router(session):
    session.install("-e", ".")
    session.install("requests", "pytest")
    session.run("pytest", "-m", "routertest")


@nox.session(python=PYTHON_TEST_VERSIONS)
def test_all(session):
    session.install("-e", ".")
    session.install("requests", "pytest")
    session.run("pytest")


@nox.session(name="check")
def ruff_check(session):
    session.run("ruff", "check", "fritzconnection", external=True)


@nox.session
def mypy(session):
    session.install("-e", ".")
    session.install("mypy==1.11.1", "types-requests", "segno")
    session.run("mypy", "fritzconnection/core/fritzconnection.py")
    session.run("mypy", "fritzconnection/core/fritzmonitor.py")
    session.run("mypy", "fritzconnection/lib")


@nox.session
def sphinx(session):
    session.install("-e", ".")
    session.install("pip-tools==7.3.0")
    session.run(
        "pip-compile",
        "--strip-extras",
        "-q",
        "--output-file=docs/requirements.out",
        "docs/requirements.txt"
        #"docs/requirements.local.in"
    )
    session.install("-r", "docs/requirements.out")
    session.run("sphinx-build", "docs", "docs_out")


@nox.session(python=PYTHON_DEVELOPMENT_VERSION)
def build(session):
    session.install("-e", ".")
    session.run("python", "setup.py", "sdist", "bdist_wheel")


@nox.session(name="check-twine", python=PYTHON_DEVELOPMENT_VERSION)
def check_twine(session):
    session.install("-e", ".")
    session.install("twine")
    session.run("twine", "check", "dist/*")


@nox.session(name="upload-to-pypi", python=PYTHON_DEVELOPMENT_VERSION)
def upload_to_pypi(session):
    session.install("-e", ".")
    session.install("twine")
    session.run("twine", "upload", "dist/*")
